#!/usr/bin/env python
# -*- coding: utf-8 -*-

#     This file is part of RAINstorm. 
#     Please, read the "Copyright" and "LICENSE" variables for copyright advisement
#   (just below this line):

COPYRIGHT = "Copyright 2010 Miguel Ángel García Martínez <miguelangel.garcia@gmail.com>"

LICENSE = """
  RAINstorm is free software; you can redistribute it and/or modify \
it under the terms of the GNU General Public License as published by \
the Free Software Foundation; either version 3 of the License, or \
(at your option) any later version.

  RAINstorm is distributed in the hope that it will be useful, \
but WITHOUT ANY WARRANTY; without even the implied warranty of \
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the \
GNU General Public License for more details.

  You should have received a copy of the GNU General Public License \
along with RAINstorm (maybe in file "COPYING"); if not, write to the Free Software \
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

# cvs version: $Id$

###################################################################################

import sys
from optparse import OptionParser
import csv
from info import info
import re

TMPL_SKELETON="""\
#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  This file has been generated by RAINstorm.
#  Please, do not touch.

$RS_TMPL_HEADERS$

class RainStorm ( object ):
   def __init__ ( self ):
      self._database = None
      self._scheme   = None

   def set_connection ( self, username, password, hostname='localhost',
                  port=0, ddbbname=None, scheme='sqlite'):
      self._database = create_database ( "%s://%s:%s@%s:%s/%s" %
            ( scheme, username, password, hostname, port, ddbbname ) )
      self._scheme = scheme

   def get_store ( self ):
      if self._database 
         return Store ( self._database )
      return None

   def get_script_sqlite ( self ):
      raise NotImplementedError ()

   def get_script_mysql ( self ):
      raise NotImplementedError ()

   def get_script_postgresql ( self ):
      raise NotImplementedError ()

   def get_script ( self, scheme=None ):
      if scheme == None:         scheme = self.scheme
      if scheme == None:         return None
      if scheme == 'myname':     return self.get_script_mysql ( self )
      if scheme == 'sqlite':     return self.get_script_sqlite ( self )
      if scheme == 'postgresql': return self.get_script_postgresql ( self )


$RS_TMPL_TABLES$
"""

TMPL_SKEL_TABLE="""
class $RS__TABLE_NAME$ (RainStorm):
   \"\"\"
   $RS_TABLE_DESCRIPTION$
   \"\"\"
   __storm_table__ = '$RS_TABLE_NAME$'
$RS_FIELD_DEFS$

   def __init__ ( self $RS_FIELD_ENUMS$ ):
      super ( $RS__TABLE_NAME$, self ).__init__ ()
$RS_FIELD_INITS$

   def get_mysql_script ( self ):
      return $RS_MYSQL_BUILDER$

   def get_sqlite_script ( self ):
      return $RS_SQLITE_BUILDER$

   def get_postgresql_script ( self ):
      return $RS_POSTGRESQL_BUILDER$
"""

TMPL_FIELD_DEFS="""\
   $RS_FIELD_NAME$ = $RS_FIELD_STORMTYPE$ ( $RS_FIELD_PRIMARY$ )
"""

TMPL_FIELD_ENUM=", $RS_FIELD_NAME$"

TMPL_FIELD_INIT="""\
      self.$RS_FIELD_NAME$ = $RS_FIELD_NAME$
"""

TMPL_MYSQL_BUILDER="""\
"""
TMPL_SQLITE_BUILDER="""\
"""
TMPL_POSTGRESQL_BUILDER="""\
"""

def replace_patterns ( origin, table ):
    """ 
    Replaces patterns in the string 'origin' by the patterns in
    the hash 'table'.
    """
    variables = sorted ( re.findall ( '\$RS_[a-zA-Z_]+\$', origin ) )
    last = ''
    retval = origin
    for v in variables:
        if last == v:
            continue
        varname = v[4:-1]
        value = ''
        if varname[0] == '_':
            value = table.get ( varname[1:], '' ).capitalize ()
        else:
            value = table.get ( varname, '' )
        retval = retval.replace ( v, value )
    return retval
                           

class Paths:
    """ Some generic paths useful for the application
    """
    is_debian_package = False
    @staticmethod
    def app_dir():
        return os.path.dirname(__file__)
    
    @staticmethod
    def locale_dir(*args):
        return os.path.join(Paths.app_dir(),"po", *args)
    
    @staticmethod
    def help_dir(*args):
        return os.path.join(Paths.app_dir(),"help", *args)
    
    @staticmethod
    def share_dir ( *args ):
        if Paths.is_debian_package:
          return os.path.join ( "/usr/share/rainstorm", *args )
        return os.path.join ( Paths.app_dir(), *args )

    @staticmethod
    def glade_dir ( *args ):
        if Paths.is_debian_package:
            return Paths.share_dir ( *args )
        return os.path.join ( Paths.app_dir(), "glade", *args )
    @staticmethod
    def config_dir(*args):
        return os.path.expanduser(os.path.join('~', '.rainstorm', *args))
    
class RainWalker:
    """
    Who knows how to build the classes from the input files.
    """
    def __init__ ( self, options, filename ):
        self.options = options
        self.filename = filename
        
        self.header = []
        self.tables = []
        self.curr_table = None

    class Table:
        """ Implements a table in BBDD. """
        def __init__ ( self, name, description ):
            self.name = name
            self.description = description
            self.fields = []
            self.references = []

        def __get_tableval ( self ):
            fielddef = ''
            fieldenum = ''
            fieldinit = ''

            fieldmysql = []
            fieldsqlite = []
            fieldpostgresql = []
    
            tableval = {}
    
            for f in self.fields:
                fielddef += f.get_definition ()
                fieldenum += f.get_enumeration ()
                fieldinit += f.get_initialization ()

                fieldmysql.append ( f.get_script_mysql () )
                fieldsqlite.append ( f.get_script_sqlite () )
                fieldpostgresql.append ( f.get_script_postgresql () )
    
            tableval ['TABLE_NAME'] = self.name
            tableval ['TABLE_DESCRIPTION'] = self.description
            tableval ['FIELD_DEFS'] = fielddef
            tableval ['FIELD_ENUMS'] = fieldenum
            tableval ['FIELD_INITS'] = fieldinit

            tableval ['MYSQL_BUILDER'] = replace_patterns ( TMPL_MYSQL_BUILDER, ', '.join (fieldmysql) )
            tableval ['SQLITE_BUILDER'] = replace_patterns ( TMPL_MYSQL_BUILDER, ', '.join (fieldsqlite) )
            tableval ['POSTGRESQL_BUILDER'] = replace_patterns ( TMPL_MYSQL_BUILDER, ', '.join (fieldpostgresql) )

            return tableval

        def get_skeleton ( self ):
            """ Gets the skeleton of the ORM """
            fieldbuild = ''
            
            return replace_patterns ( TMPL_SKEL_TABLE, self.__get_tableval () )

    class Field:
        """ Implements a Field inside a Table """
        TYPES_postgresql = {
            'Bool'     : 'BOOL',
            'Int'      : 'INT',
            'Float'    : 'FLOAT',
            'Decimal'  : 'NUMERIC',
            'Unicode'  : 'TEXT',
            'RawStr'   : 'BYTEA',
            'Pickle'   : 'BYTEA',
            'DateTime' : 'TIMESTAMP',
            'Date'     : 'DATE',
            'Time'     : 'TIME',
            'TimeDelta': 'INTERVAL',
            }
        TYPES_mysql = {
            'Bool'     : 'TINYINT(1)',
            'Int'      : 'INT',
            'Float'    : 'FLOAT',
            'Decimal'  : 'NUMERIC',
            'Unicode'  : 'TEXT',
            'RawStr'   : 'BINARY',
            'Pickle'   : 'BINARY',
            'DateTime' : 'TIMESTAMP',
            'Date'     : 'DATE',
            'Time'     : 'TIME',
            'TimeDelta': 'TEXT',
            }
        TYPES_sqlite = {
            'Bool'     : 'INT',
            'Int'      : 'INT',
            'Float'    : 'FLOAT',
            'Decimal'  : 'TEXT',
            'Unicode'  : 'TEXT',
            'RawStr'   : 'BLOB',
            'Pickle'   : 'BLOB',
            'DateTime' : 'TEXT',
            'Date'     : 'TEXT',
            'Time'     : 'TEXT',
            'TimeDelta': 'TEXT',
            }
        def __init__ ( self, name, ftype, description ):
            name = name.strip().lower()
            primary = False
            if name[0] == '*':
                name = name[1:].strip()
                primary = True
            self._tableval = self.__get_values ( name, ftype, description, primary )
            
        def __get_values ( self, name, ftype, description, primary ):
            tableval = {}
            tableval ['FIELD_NAME'] = name.lower()
            tableval ['FIELD_STORMTYPE'] = ftype
            if primary:
                tableval ['FIELD_PRIMARY'] = 'primary=True'
                tableval ['MYSQL_PRIMARY'] = 'PRIMARY KEY'
                tableval ['SQLITE_PRIMARY'] = 'PRIMARY KEY'
                tableval ['POSTGRESQL_PRIMARY'] = 'PRIMARY KEY'
            tableval ['MYSQL_TYPE'] = self.TYPES_mysql.get ( ftype, None )
            tableval ['SQLITE_TYPE'] = self.TYPES_sqlite.get ( ftype, None )
            tableval ['POSTGRESQL_TYPE'] = self.TYPES_postgresql.get ( ftype, None )
            return tableval

        def get_definition ( self ):
            return replace_patterns ( TMPL_FIELD_DEFS, self._tableval )
            
        def get_initialization ( self ):
            return replace_patterns ( TMPL_FIELD_INIT, self._tableval )
            
        def get_enumeration ( self ):
            return replace_patterns ( TMPL_FIELD_ENUM, self._tableval )

        def get_script_sqlite ( self ):
            tmpl = "$RS_FIELD_NAME$ $RS_SQLITE_TYPE$ $RS_SQLITE_PRIMARY$"
            return replace_patterns ( tmpl, self._tableval )

        def get_script_mysql ( self ) :
            tmpl = "$RS_FIELD_NAME$ $RS_MYSQL_TYPE$ $RS_MYSQL_PRIMARY$"
            return replace_patterns ( tmpl, self._tableval )

        def get_script_postgresql ( self ) :
            tmpl = "$RS_FIELD_NAME$ $RS_POSTGRESQL_TYPE$ $RS_POSTGRESQL_PRIMARY$"
            return replace_patterns ( tmpl, self._tableval )
            
   
    class Reference:
        def __init__ ( self, field, reftable, reffield ):
            self.field = field
            self.reftable = reftable
            self.reffield = reffield

    def add_header ( self, head ):
        self.header.append ( head )

    def add_table ( self, name, description = '' ):
        self.curr_table = self.Table ( name, description )
        self.tables.append ( self.curr_table )

    def add_field ( self, name, kind, description='' ):
        field = self.Field ( name, kind, description )
        self.curr_table.fields.append ( field )

    def add_reference ( self, field_name, ref_table, ref_field ):
        ref = self.Reference ( field_name, ref_table, ref_field )
        self.curr_table.references.append ( ref )

    #abstractmethod
    def run ( self ):
        raise NotImplementedError ()

    def __str__ ( self ):
        retval = ''

        for h in self.header:
            retval += '%s\n' % (h)
        for t in self.tables:
            retval += 't = ' + t.name

        return retval
    
    def generate_bbddscript ( self ):
        tableval = {}
        tables = ''
        for t in self.tables:
            tables += t.get_skeleton ()
        tableval ['TMPL_HEADERS'] = '\n'.join ( self.header )
        tableval ['TMPL_TABLES']  = tables

        return replace_patterns ( TMPL_SKELETON, tableval )


class CsvWalker ( RainWalker ):

    def run ( self ):
        filed = open ( self.filename )
        reader = csv.reader ( filed, delimiter=',', quotechar='|' )
        for row in reader:
            if len ( row ) == 0:
                continue
            ind = row[0].strip().lower()
            if ind in ['h', 'head', 'header']:
                self.add_header ( ''.join (row[1:]) )
            elif ind in ['t', 'table']:
                self.add_table ( * row[1:3] )
            elif ind in ['f', 'field']:
                self.add_field ( * row[1:3] )
            elif ind in ['r', 'reference']:
                self.add_reference ( * row [1:4] )
        

        

if __name__ == "__main__":
    parser = OptionParser ( usage="usage: %prog [options]",
        version = "    %s %s by %s %s" % ("%prog",
                info.version, info.author, info.author_email),
        description = "%s.\nBugs, queries or improvements refer %s" % (
                info.description, info.url ),
        )
    #parser.add_option ( "", "" )
    options, files = parser.parse_args ()
    if len ( files ) == 0:
        print "Sorry, you need to ask for a file."
    elif len ( files ) > 1:
        print "Sorry, you need to ask for ONLY ONE file."
    else:
        print "Running..."
        c = CsvWalker ( options, '%s'%files[0] )
        c.run ()
        print c.generate_bbddscript ()
